# -*- coding: utf-8 -*-
import sys
import pkgutil
import importlib
import inspect
from datetime import datetime


def pycallable(func):
    if sys.version_info.major == 2:
        return callable(func)
    else:
        return hasattr(func, '__call__')

def compact(*names):
    """
    打包变量为字典
    :param names:
    :return:
    """
    caller = inspect.stack()[1][0]  # caller of compact()
    vars = {}
    for n in names:
        if n in caller.f_locals:
            vars[n] = caller.f_locals[n]
        elif n in caller.f_globals:
            vars[n] = caller.f_globals[n]
    return vars


def extract(vars):
    """
    提取字典为变量
    :param vars:
    :return:
    """
    caller = inspect.stack()[1][0]  # caller of extract()
    for n, v in vars.items():
        caller.f_locals[n] = v  # NEVER DO THIS - not guaranteed to work


# 可转化字符串类型的数字为int
def isint_or(var, replace=0):
    """
    判断是否int, 否则使用replace替换
    :param var:
    :param replace:
    :return:
    """
    if sys.version_info.major == 2:
        # python2
        if isinstance(var, int) or isinstance(var, long):
            return var
        if isinstance(var, (str, unicode)) and var.isdigit():
            return int(var)
        return replace
    else:
        # python3
        if isinstance(var, int):
            return var
        if isinstance(var, str) and var.isdigit():
            return int(var)
        return replace


# 拼接数据为字符串
# 当把一组ID用逗号分割拼接为字符串时, 使用 str.join()会报错
# 故需要一个拼接方法, 自适应数据类型拼接成字符串
def implode(sep, items):
    """
    拼接数据为字符串
    :param sep: 间隔字符
    :param items: 数据元组
    :return: string, 拼接后的字符串

    nums = [1,2,3,4,5]
    nums = implode(":", nums) // >> "1:2:3:4:5"

    """
    return str(sep).join(map(__stringify, items))


def splitids(str):
    """
    分离id串为int数组
    :para str: id串
    :return:   list(ids)
    """
    str = str.strip()
    if not str:
        return []
    return list(map(lambda id: int(id), str.split(",")))


def __stringify(v):
    """
    implode 转化数据为字符串回调方法
    """
    if v is None:
        return ''
    if sys.version_info.major == 2:
        # python2
        if isinstance(v, (str, unicode)) and len(v) == 0:
            return ''
        if isinstance(v, (str, unicode)):
            return v
        return "%s" % (v,)
    else:
        # python3
        if isinstance(v, str) and len(v) == 0:
            return ''
        if isinstance(v, str):
            return v
        return "%s" % (v,)


def import_submodules(package_name):
    """ Import all submodules of a module, recursively

    :param package_name: Package name
    :type package_name: str
    :rtype: dict[types.ModuleType]
    """
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + '.' + name)
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)
        }


def list_or_args(keys, args):
    # returns a single list combining keys and args
    try:
        iter(keys)
        # a string or bytes instance can be iterated, but indicates
        # keys wasn't passed as a list
        if isinstance(keys, (str, bytes)):
            keys = [keys]
        if len(args):
            keys = [keys]
    except TypeError:
        keys = [keys]
    if args:
        keys.extend(args)
    return keys


def strftime(time=None):
    # 返回SQL时间串
    if not time:
        time = datetime.now()
    return time.strftime("%Y-%m-%d %H:%M:%S")


def strptime(timestr):
    return datetime.strptime(timestr, "%Y-%m-%d %H:%M:%S")


def isstr(val):
    if sys.version_info.major == 2:
        return isinstance(val, (str, unicode))
    else:
        return isinstance(val, str)



# -*- coding: utf-8 -*-
from future.utils import raise_from
import sys
from redis.client import list_or_args
from redis.exceptions import RedisError
from sqlalchemy import event
from sqlalchemy import inspect
from .. import JSON, implode
import time, datetime
from inspect import isclass, ismethod, isfunction, getargspec, getsource
from flask import current_app as app
from .application import db, get_db_engine, redis

import dill
from ..exceptions import *
from .. import isstr

cache_life_time = app.config.get('CACHE_LIFE_TIME', 600)


# 条件混合
def cache_slug(*args):
    return ":".join(map(lambda x: x if isinstance(x, str) else str(x), args))


class ModelCache(object):
    """缓存对象"""
    name = ""
    __redis = None

    __caches = {}

    @classmethod
    def fetch(cls, name):
        if name in cls.__caches:
            return cls.__caches[name]
        cache = cls(name)
        return cache

    def __init__(self, name):
        self.name = name
        self.__redis = redis
        self.__class__.__caches[name] = self

    def serialize(self, val):
        return dill.dumps(val)

    def unserialize(self, val):
        return dill.loads(val) if val else None

    def handle(self):
        return self.__redis

    # 清空类相关缓存
    def clear(self, match=None, count=5000, cursor=0):
        num = 0
        if match:
            rmatch = "%s#%s" % (self.name, match)
        else:
            rmatch = "%s#*" % (self.name,)
        cursor, keys = self.__redis.scan(cursor=cursor, match=rmatch, count=count)
        if len(keys):
            num += self.__redis.delete(*keys)
        if cursor != 0:     # 还需要遍历
            num += self.clear(match, count, cursor)
        return num

    def rkey(self, key):
        return "%s#%s" % (self.name, key)

    # redis set
    def set(self, key, value, ex=None):
        return self.__redis.set(self.rkey(key), self.serialize(value), ex)

    # redis mset
    def mset(self, *args, **kwargs):
        # if args:
        #     if len(args) != 1 or not isinstance(args[0], dict):
        #         raise RedisError('MSET requires **kwargs or a single dict arg')
        #     kwargs.update(args[0])
        pipe = self.__redis.pipeline()
        for k, v in args:
            pipe.set(self.rkey(k), self.serialize(v), cache_life_time)
        results = pipe.execute()
        return results[-1]

    # redis get
    def get(self, key):
        return self.unserialize(self.__redis.get(self.rkey(key)))

    # redis mget
    def mget(self, keys, *args):
        keys = list_or_args(keys, args)
        gkeys = map(lambda k: self.rkey(k), keys)
        gets = self.__redis.mget(*gkeys)
        return map(lambda v: self.unserialize(v), gets)

    # @staticmethod
    # def try_expired(d):
    #     if d is None:
    #         return False
    #     if not d or not isinstance(d, dict):
    #         return True
    #     expired = d.get("e", 0)
    #     if not expired:
    #         return True
    #     ts = time.mktime(datetime.datetime.now().timetuple())
    #     if expired and expired <= ts:
    #         return True
    #     return False

    # @staticmethod
    # def set_data(value, ex=None):
    #     return {
    #         "e": (time.mktime(datetime.datetime.now().timetuple()) + ex) if (ex and isinstance(ex, int)) else 0,
    #         "d": value
    #     }

    # def get_data(self, key, unserialized):
    #     if unserialized is None:
    #         return None
    #     if self.try_expired(unserialized):
    #         self.delete(key)
    #         return None
    #     if isinstance(unserialized, dict):
    #         return unserialized.get("d", None)
    #     return None

    # redis delete
    def delete(self, keys, *args):
        keys = list_or_args(keys, args)
        dkeys = map(lambda k:self.rkey(k), keys)
        return self.__redis.delete(*dkeys)


class CacheableBase(db.Model):
    """
        通用缓存模型

        class TestModel(CacheableBase):
            __tablename__ = 'test'

    """
    # 抽象模型
    __abstract__ = True
    __session__ = db.session
    # 根据字段作缓存
    __cached_fields__ = []
    # 标记auto_cached
    __auto_cached__ = False

    # stored with ModelCache instance
    __cache__ = None

    # default expires in 10 mimutes
    __cache_expires__ = cache_life_time

    @classmethod
    def _cache(cls, subfix="", noconflict=False):
        if cls.__cache__:
            return cls.__cache__
        path = [cls.__module__, cls.__name__]
        if subfix: path.append(subfix)
        path = ".".join(path)
        cls.__cache__ = ModelCache(path if noconflict else cls.__name__)
        return cls.__cache__

    @classmethod
    def get(cls, ident, flush=False):
        """pk取单个数据实例"""
        if not ident:
            return None
        cache = cls._cache()
        slug = cache_slug("PK", ident)
        if not flush:
            cached = cache.get(slug)
            if cached:
                if not hasattr(cls, "__session__"):
                    raise ReferenceError("__session__ haven't bind to %s" % cls.__name__)
                if inspect(cached).persistent:
                    return cached
                return cls.__session__.merge(cached)

        # 未缓存
        record = None
        if cls.query:
            record = cls.query.get(ident)
            if record:
                cache.set(slug, record, ex=cls.__cache_expires__)
        return record

    @classmethod
    def mget(cls, pks=[], *args):
        """
        pk批量取数据实例

        Model.mget(1,2,5)
        Model.mget([1,2,4,5])
        @:returns [instance1, instance2, instance3, ..., None ]
        """
        cache = cls._cache()
        # [ id1, id2, id3, ... ]
        pks = list_or_args(pks, args)
        if not len(pks):
            return []
        pslugs = map(lambda k: cache_slug("PK", str(k)), pks)
        # 先尝试缓存命中数据
        caches = cache.mget(pslugs)
        rect = {}
        lost = []
        for i, cached in enumerate(caches, start=0):
            if cached is None:
                lost.append(pks[i])
            else:
                if not hasattr(cls, "__session__"):
                    raise ReferenceError("__session__ haven't bind to %s" % cls.__name__)
                rect[str(pks[i])] = cached if inspect(cached).persistent else cls.__session__.merge(cached)

        # 未命中的从数据库中取
        pk = cls._pk()
        if len(lost):
            records = cls.query.filter(getattr(cls, pk).in_(lost)).all()
            for record in records:
                rect[str(getattr(record, pk))] = record
            # 批量缓存
            setcaches = map(lambda pkid: (cache_slug("PK", pkid), rect.get(str(pkid), None)), lost)
            setcaches = list(filter(lambda c: c[1], setcaches))
            if len(setcaches):
                cache.mset(*setcaches)

        # 按传入的参数顺序返回数据
        return list(map(lambda k: rect[str(k)] if str(k) in rect else None, pks))

    @classmethod
    def _pk(cls):
        """first primary key"""
        detection = inspect(cls)
        pks = detection.primary_key
        return pks[0].name if len(pks) else None

    @classmethod
    def save(cls, *args, **kwargs):
        instance = cls.__save(*args, **kwargs)
        cls.__session__.commit()
        return instance

    @classmethod
    def flush_save(cls, *args, **kwargs):
        instance = cls.__save(*args, **kwargs)
        cls.__session__.flush()
        cls.__session__.refresh(instance)
        return instance

    @classmethod
    def __save(cls, *args, **kwargs):
        """
        保存/修改模型数据
        :param args:
        :param kwargs:
        :return:
        """
        isnew = False
        pk = cls._pk()
        pkid = None
        if pk in kwargs:      # 存在id参数
            pkid = kwargs[pk]
            del kwargs[pk]    # 移除id参数
            if isstr(pkid) and pkid.isdigit():
                pkid = int(pkid)
            if not pkid:
                isnew = True    # id参数不规范, 视为新创建
        else:
            isnew = True
        """ 需要验证cls()是否支持hybird属性 """
        if isnew:
            instance = cls()
        else:
            instance = cls.get(pkid)

        # 传了PK, 但记录不存在, 新建记录
        if not instance:
            instance = cls()
            if pkid is not None:
                kwargs[pk] = pkid

        # 批量设置属性
        instance.setattrs(**kwargs)
        cls.__session__.add(instance)
        return instance

    @classmethod
    def delete(cls, pkid):
        """
        删除记录
        :param pkid, 主键值
        """
        # remove db record
        record = cls.get(pkid)
        if record:
            cls.__session__.delete(record)
            # commit
            cls.__session__.commit()
            return True
        else:
            return False

    @classmethod
    def emptyInstance(cls):
        return cls()

    @classmethod
    def auto_increment(cls):
        """
        获取表的auto_increment
        :return:
        """
        result = db.session.execute(
            "SELECT AUTO_INCREMENT FROM information_schema.tables WHERE table_name='%s'" % cls.__tablename__,
            bind=get_db_engine(cls.__bind_key__))
        rows = result.fetchall()
        return rows[0][0] if len(rows) else None


    def setattrs(self, *args, **kwargs):
        """
        批量设置属性
        :param args:
        :param kwargs:
        :return:
        """
        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise AttributeError('set_attrs requires **kwargs or a single dict arg')
            kwargs.update(args[0])

        for k, v in kwargs.items():
            if hasattr(self, k):  # 类有该属性
                setattr(self, k, v)
        return self

    def flush_cache(self):
        """清除实例缓存"""
        cls = self.__class__
        cache = cls._cache()
        pk = cls._pk()
        if hasattr(self, pk):
            cache.delete(cache_slug("PK", getattr(self, pk)))
        pass

    @classmethod
    def flush_pk(cls, pkid):
        cache = cls._cache()
        cache.delete(cache_slug("PK", pkid))

    @classmethod
    def flush_all(cls):
        """清空所有类下缓存"""
        return cls._cache().clear()

    def to_json(self):
        """
        转化实例对象为JSON字符串

        :return:
        """
        return JSON.stringify(self)

    def __json__(self):
        """
        默认JSON序列化回调
        :return:
        """
        dict = {}
        for k, v in self.__dict__.iteritems():
            # 过滤下划线的属性
            if k.find("_", 0, 1) == 0:
                continue
            dict[k] = v
        return dict

    @classmethod
    def __modified__(cls, mapper, connection, target):
        """
        修改记录时调用

        :return:
        """
        return True

    @classmethod
    def auto_cached(cls):
        """
        开启自动缓存

        Model.auto_cached()
        """
        if cls.__auto_cached__:
            return

        cls.__auto_cached__ = True

        @event.listens_for(cls, 'after_update')
        def after_update_handel(mapper, connection, target):
            """
            更新缓存
            update, insert, delete 一条记录时, 都会触发该事件
            :param mapper:
            :param connection:
            :param target:
            :return:
            """
            target.flush_cache()
            if cls.__modified__(mapper, connection, target) is False:
                return

        @event.listens_for(cls, 'after_insert')
        def before_insert_handel(mapper, connection, target):
            """
            插入数据前更新缓存
            :param mapper:
            :param connection:
            :param target:
            :return:
            """
            target.flush_cache()
            if cls.__modified__(mapper, connection, target) is False:
                return

        @event.listens_for(cls, 'after_delete')
        def after_delete_handel(mapper, connection, target):
            """
            删除数据后清缓存

            :param mapper:
            :param connection:
            :param target:
            :return:
            """
            target.flush_cache()
            if cls.__modified__(mapper, connection, target) is False:
                return


def cacheable(expire=cache_life_time, noconflict=False):
    _ = {
        "method": None
    }

    def init(method, *args, **kwargs):
        _['method'] = method
        m.func_name = method.__name__
        m._cache_fun = method
        return m

    def m(*args, **kwargs):
        cls = None
        if len(args) > 0:
            a0 = args[0]
            if isclass(a0):
                cls = a0
            elif isclass(a0.__class__):
                cls = a0.__class__
        method = _['method']
        path = slugpath(method, cls, noconflict)
        slug = argslug(method, False,  *args, **kwargs)
        cache = ModelCache.fetch(path)
        cached = cache.get(slug)
        if cached is not None:
            if cls and hasattr(cls, "__session__"):
                if isinstance(cached, (tuple, list, set)):
                    for c in cached:
                        if isinstance(c, CacheableBase) and c not in cls.__session__:
                            cls.__session__.merge(c, load=False)
                if isinstance(cached, CacheableBase) and cached not in cls.__session__:
                    cls.__session__.merge(cached, load=False)
            return cached
        # 原接口取数据
        data = method(*args, **kwargs)
        cache.set(slug, data, expire)
        return data

    return init


def flushcache(method, *args, **kwargs):
    fun = method._cache_fun if hasattr(method, "_cache_fun") else None
    if fun:
        cls = None
        # 清空指定参数缓存
        args = list(args)
        if ismethod(method):
            if sys.version_info.major == 2:
                cls = method.im_self    # python2
            else:
                cls = method.__self__   # python3
        path = slugpath(fun, cls)
        cache = ModelCache.fetch(path)
        if not len(args) and not len(kwargs):
            # 清空方法缓存
            cache.clear()
            return True

        if cls:
            args.insert(0, cls)
        slug = argslug(fun, True, *args, **kwargs)
        if slug:
            cache.clear(slug)
        return True
    else:
        return False


def slugpath(method, cls=None, noconflict=False):
    path = "Function"
    if cls:
        path = implode(".", [cls.__module__, cls.__name__]) if noconflict else cls.__name__
    path = implode(".", [path, method.__name__])
    return path


def argslug(method, wildcard=False, *args, **kwargs):
    """
    获取参数slug
    :param method
    :param wildcard: 是否通配符模式
    """
    argspec = getargspec(method)
    argnames = argspec.args if argspec.args else () # tuple
    argdef_list = argspec.defaults if argspec.defaults else () # tuple
    argdefs = dict()    # 参数默认值, 字典, 保存有默认值的参数名和值
    defindex = len(argnames) - len(argdef_list)
    for i, defv in enumerate(argdef_list):
        # 绑定默认值
        argdefs[argnames[defindex+i]] = defv

    if len(argnames):
        # 分离第一个参数 @classmethod
        if argnames[0] in kwargs:
            if isclass(kwargs[argnames[0]]):
                kwargs.pop(argnames[0])
                argnames = argnames[1:]

        else:
            if len(args) and isclass(args[0]):
                args = args[1:]
                argnames = argnames[1:]

    slugs = list()
    for i, name in enumerate(argnames):
        v = "*"
        has_value = False
        if name in kwargs:
            v = kwargs.pop(name)
            has_value = True
        if not has_value and len(args) > i:
            v = args[i]
            has_value = True
        if not wildcard and not has_value and (name in argdefs):
            v = argdefs.get(name)
            has_value = True
        if not wildcard and not has_value:
            continue
        slugs.append("%s=%s" % (name, v))

    if len(args) > len(argnames):
        args = args[len(argnames):]
    else:
        args = []
    for v in args:
        slugs.append("%s" % v)

    kws = kwargs.items()
    kws = sorted(kws, key=lambda kv: kv[0], reverse=False)
    for k, v in kws:
        slugs.append("%s=%s" % (k, v))
    slugs = implode(",", slugs)
    return "("+slugs+")" if slugs else "()"



# -*- coding: utf-8 -*-
from wesell.flask.cacheable import CacheableBase

from sqlalchemy import Column, Integer, String, BIGINT, TIMESTAMP, DATE, Enum, Boolean, event, or_, not_, Text, Table
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.attributes import InstrumentedAttribute, get_history
from datetime import datetime
from wesell.flask import operator_id as operator_id_handle, operator_role as operator_role_handle, \
    operator_name as operator_name_handle, app
from .application import db
from ..sqlalchemy.types import JSONSerializer
from .. import JSON
from .. import compact, extract, pycallable
from functools import wraps
from flask import render_template_string
from sqlalchemy import event, inspect
from flask_rq import job
import re
from numbers import Number


class SqllogBase(db.Model):
    """基本日志模型
    """
    __sql_log__ = "0.1.0"
    __abstract__ = True
    __bind_model__ = None
    __bind_key__ = app.config.get("SQLLOG_BIND_KEY", None)
    __tablename__ = app.config.get("SQLLOG_TABLENAME", "sqllog")
    __session__ = db.session
    # 监听的SQL事件
    listen = ("insert", "update", "delete")
    # 不记录属性变化数据(建议忽略大数据的字段)
    truncated_fields = tuple()

    __message_default_templates__ = dict(
        insert="{operator}创建了{model}#{target_id}",
        update="{operator}更新了{model}#{target_id}：{fields sp=，}{key}=>{after}{/fields}",
        delete="{operator}删除了{model}#{target_id}"
    )
    # 简易消息模板配置, 未设置会使用默认配置 __message_default_templates__
    # 模板变量:
    #   operator:       优先显示操作者的名字或 {role}#{id}
    #   operator_id:    操作者ID(未设置为0)
    #   operator_role:  操作者角色(如有,如Admin, 默认为空)
    #   operator_name:  操作者名称, 默认为"unset"
    #   model:  监听的模型类名
    #   target_id: 模型实例ID
    #
    #   仅update时:
    #   {fields sp=分割符}遍历变更字段{/fields}
    #   key: 字段名
    #   before: 变更前值(json)
    #   after: 变更后值(json)
    #
    message_templates = dict()

    # Table fields
    id = Column(BIGINT, primary_key=True)
    message = Column(Text, doc="简易消息文本日志")
    action = Column(Enum("insert", "update", "delete"), default="update", server_default="update", doc="日志类型")
    model = Column(String(32), index=True, default="", server_default="", doc="模型名称")
    target_id = Column(BIGINT, default="0", server_default="0", doc="模型实例的ID")
    operator_id = Column(BIGINT, default=operator_id_handle, server_default="0", doc="操作人id")
    operator_role = Column(String(32), default=operator_role_handle, server_default="", doc="操作人角色")
    operator_name = Column(String(32), default="", server_default="", doc="操作人名")
    # json数据, 保存发生变化的属性数据
    # {
    #   "key1":[before, after],
    #   "key2":[before, after]
    # }
    changes = Column(JSONSerializer, doc="字段变更详情")
    # 创建时间
    ctime = Column(TIMESTAMP, default=datetime.now, doc='创建时间, 默认为当前时间')

    def __init__(self, action, target):
        cls = self.__class__
        model_cls = target.__class__
        pk = get_pk(model_cls)
        self.action = action
        self.model = model_cls.__name__
        self.target_id = getattr(target, pk, 0)
        operator = cls.operator_handle(target, action)
        self.operator_id = operator[0]
        self.operator_role = operator[1]
        self.operator_name = operator[2]
        self.changes = cls.__get_changes(target)
        self.message = cls.__render_message(self)

    @classmethod
    def __render_message(cls, log):
        operator_id = log.operator_id
        operator_role = log.operator_role
        operator_name = log.operator_name
        action = log.action
        model = log.model
        target_id = log.target_id
        operator = operator_name
        changes = log.changes
        if operator_name == "System" and (operator_role or operator_id):
            operator = "%s#%s" % (operator_role, operator_id)
        # 模块配置
        templates = cls.__message_default_templates__.copy()
        templates.update(cls.message_templates)
        template = templates.get(action, "模板获取错误")
        if pycallable(template):
            template = template(log)
        contexts = compact("operator_id", "operator_role", "operator_name", "operator", "action", "model", "target_id")
        # update fields
        if action == "update":
            field_re = re.compile('\{fields sp=(.*?)\}(.*)\{/fields\}', flags=re.DOTALL | re.I | re.U)
            field_finds = field_re.findall(template)
            if field_finds:
                template = field_re.sub("{fields}", template)
                sp = field_finds[0][0]
                tpl = field_finds[0][1]
                chstr = []
                for k, v in log.changes.items():
                    kstr = tpl.replace("{key}", k).replace("{before}", JSON.stringify(v[0])).\
                        replace("{after}", JSON.stringify(v[1]))
                    chstr.append(kstr.encode("utf-8"))
                contexts.setdefault("fields", sp.join(chstr))

        for k, v in contexts.items():
            replace = re.compile('\{%s\}' % k, flags=re.U)
            if isinstance(v, unicode):
                v = v.encode("utf-8")
            template = replace.sub("%s" % v, template)
        return template

    @classmethod
    def __get_changes(cls, target):
        changes = dict()
        model = target.__class__
        for k, v in model.__dict__.items():
            if isinstance(v, InstrumentedAttribute):
                his = get_history(target, k)
                if his.deleted:  # 有删除的数据
                    if k in cls.truncated_fields:
                        changes.setdefault(k, ["[truncated]", "[truncated]"])
                    else:
                        if his.deleted[0] == his.added[0]:
                            continue
                        if isinstance(his.deleted[0], Number) and his.deleted[0] == float(his.added[0]):
                            continue
                        changes.setdefault(k, [his.deleted[0], his.added[0]])
                continue
        return changes

    @classmethod
    def operator_handle(cls, target, action, **kwargs):
        return (operator_id_handle(), operator_role_handle(), operator_name_handle() or "System")

    @classmethod
    def bind_model(cls, model):
        if cls.__bind_model__:
            if cls.__bind_model__ != model:
                raise Exception("每个SqllogBase类只能绑定一个模型类!")
            return  # 已经绑定过了
        cls.__bind_model__ = model
        listen_log(cls)

    @classmethod
    def handle_log(cls, log):
        pass


def listen_log(loger):
    """
    监听模型数据变动
    :param loger:
    :param listen:
    :param fields:
    :return:
    """
    # 获取绑定的数据模型
    model = loger.__bind_model__

    @event.listens_for(model, 'after_update')
    def after_update_handel(mapper, connection, target):
        if "update" in loger.listen:
            loger.handle_log(loger(action="update", target=target))

    @event.listens_for(model, 'after_insert')
    def after_insert_handel(mapper, connection, target):
        if "insert" in loger.listen:
            loger.handle_log(loger(action="insert", target=target))

    @event.listens_for(model, 'after_delete')
    def after_delete_handel(mapper, connection, target):
        if "delete" in loger.listen:
            loger.handle_log(loger(action="delete", target=target))


def get_pk(cls):
    if getattr(cls, "__pk__", None):
        return cls.__pk__
    detection = inspect(cls)
    pks = detection.primary_key
    pk = pks[0].name if len(pks) else None
    cls.__pk__ = pk
    return pk


# def make_log(loger, action, target):
#     log = loger(action=action, target=target)
#     db.session.add(log)
#     db.session.commit()
#



# -*- coding: utf-8 -*-

from functools import partial
from werkzeug.local import LocalStack, LocalProxy
from flask_redis import FlaskRedis
from flask_sqlalchemy import SQLAlchemy
from flask_rq import RQ
from celery import Celery
from flask_sqlalchemy import _BoundDeclarativeMeta as DefaultMeta, declarative_base, _QueryProperty, Model as BaseModel

redis = FlaskRedis()
db = SQLAlchemy(session_options={"autoflush":False})
celery = Celery()
rq = RQ()


class ModelMeta(DefaultMeta):
    def __new__(cls, name, bases, d):
        return DefaultMeta.__new__(cls, name, bases, d)

    def __init__(cls, name, bases, d):
        DefaultMeta.__init__(cls, name,bases, d)
        # 检测Sqllog字段, 绑定日志配置
        for k, v in d.items():
            if isinstance(v, object) and getattr(v, "__sql_log__", None):
                setattr(cls, k, v.bind_model(cls))
                break
        # auto_cache特性
        # 暂无法找到方法实现自动

# The following copied from how flask_sqlalchemy creates it's Model
Model = declarative_base(cls=BaseModel, name='Model', metaclass=ModelMeta)
Model.query = _QueryProperty(db)


# Need to replace the original Model in flask_sqlalchemy, otherwise it
# uses the old one, while you use the new one, and tables aren't
# shared between them
db.Model = Model


stacker = {
    "app": None
}

def _lookup_(name):
    return stacker[name]

app=LocalProxy(partial(_lookup_,"app"))


# -*- coding: utf-8 -*-
from sqlalchemy.types import UserDefinedType, Text
from .. import JSON


class JSONSerializer(Text):
    """JSON序列化数据格式.

    """
    __visit_name__ = 'text'

    def bind_processor(self, dialect):
        def process(value):
            return JSON.stringify(value)

        return process

    def result_processor(self, dialect, coltype):
        def process(value):
            return JSON.parse(value)

        return process



# -*- coding: utf-8 -*-
"""
    flask-mail修改版
    flask-mail: http://www.pythondoc.com/flask-mail/
    修改支持多邮件帐户
"""

from __future__ import with_statement

__version__ = '0.9.1'

import re
import blinker
import smtplib
import sys
import time
import unicodedata

from email import charset
from email.encoders import encode_base64
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.header import Header
from email.utils import formatdate, formataddr, make_msgid, parseaddr
from contextlib import contextmanager

from flask import current_app

PY3 = sys.version_info[0] == 3

PY34 = PY3 and sys.version_info[1] >= 4

if PY3:
    string_types = str,
    text_type = str
    from email import policy
    message_policy = policy.SMTP
else:
    string_types = basestring,
    text_type = unicode
    message_policy = None

charset.add_charset('utf-8', charset.SHORTEST, None, 'utf-8')


class FlaskMailUnicodeDecodeError(UnicodeDecodeError):
    def __init__(self, obj, *args):
        self.obj = obj
        UnicodeDecodeError.__init__(self, *args)

    def __str__(self):
        original = UnicodeDecodeError.__str__(self)
        return '%s. You passed in %r (%s)' % (original, self.obj, type(self.obj))


def force_text(s, encoding='utf-8', errors='strict'):
    """
    Similar to smart_text, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    if isinstance(s, text_type):
        return s

    try:
        if not isinstance(s, string_types):
            if PY3:
                if isinstance(s, bytes):
                    s = text_type(s, encoding, errors)
                else:
                    s = text_type(s)
            elif hasattr(s, '__unicode__'):
                s = s.__unicode__()
            else:
                s = text_type(bytes(s), encoding, errors)
        else:
            s = s.decode(encoding, errors)
    except UnicodeDecodeError as e:
        if not isinstance(s, Exception):
            raise FlaskMailUnicodeDecodeError(s, *e.args)
        else:
            s = ' '.join([force_text(arg, encoding, errors) for arg in s])
    return s

def sanitize_subject(subject, encoding='utf-8'):
    try:
        subject.encode('ascii')
    except UnicodeEncodeError:
        try:
            subject = Header(subject, encoding).encode()
        except UnicodeEncodeError:
            subject = Header(subject, 'utf-8').encode()
    return subject

def sanitize_address(addr, encoding='utf-8'):
    if isinstance(addr, string_types):
        addr = parseaddr(force_text(addr))
    nm, addr = addr

    try:
        nm = Header(nm, encoding).encode()
    except UnicodeEncodeError:
        nm = Header(nm, 'utf-8').encode()
    try:
        addr.encode('ascii')
    except UnicodeEncodeError:  # IDN
        if '@' in addr:
            localpart, domain = addr.split('@', 1)
            localpart = str(Header(localpart, encoding))
            domain = domain.encode('idna').decode('ascii')
            addr = '@'.join([localpart, domain])
        else:
            addr = Header(addr, encoding).encode()
    return formataddr((nm, addr))


def sanitize_addresses(addresses, encoding='utf-8'):
    return list(map(lambda e: sanitize_address(e, encoding), addresses))


def _has_newline(line):
    """Used by has_bad_header to check for \\r or \\n"""
    if line and ('\r' in line or '\n' in line):
        return True
    return False

class Connection(object):
    """Handles connection to host."""

    def __init__(self, mail):
        self.mail = mail

    def __enter__(self):
        if self.mail.suppress:
            self.host = None
        else:
            self.host = self.configure_host()

        self.num_emails = 0

        return self

    def __exit__(self, exc_type, exc_value, tb):
        if self.host:
            self.host.quit()

    def configure_host(self):
        if self.mail.use_ssl:
            host = smtplib.SMTP_SSL(self.mail.server, self.mail.port)
        else:
            host = smtplib.SMTP(self.mail.server, self.mail.port)

        host.set_debuglevel(int(self.mail.debug))

        if self.mail.use_tls:
            host.starttls()
        if self.mail.username and self.mail.password:
            host.login(self.mail.username, self.mail.password)

        return host

    def send(self, message, envelope_from=None):
        """Verifies and sends message.

        :param message: Message instance.
        :param envelope_from: Email address to be used in MAIL FROM command.
        """
        assert message.send_to, "No recipients have been added"

        assert message.sender, (
                "The message does not specify a sender and a default sender "
                "has not been configured")

        if message.has_bad_headers():
            raise BadHeaderError

        if message.date is None:
            message.date = time.time()

        if self.host:
            self.host.sendmail(sanitize_address(envelope_from or message.sender),
                               list(sanitize_addresses(message.send_to)),
                               message.as_bytes() if PY3 else message.as_string(),
                               message.mail_options,
                               message.rcpt_options)

        email_dispatched.send(message, app=current_app._get_current_object())

        self.num_emails += 1

        if self.num_emails == self.mail.max_emails:
            self.num_emails = 0
            if self.host:
                self.host.quit()
                self.host = self.configure_host()

    def send_message(self, *args, **kwargs):
        """Shortcut for send(msg).

        Takes same arguments as Message constructor.

        :versionadded: 0.3.5
        """

        self.send(Message(*args, **kwargs))


class BadHeaderError(Exception):
    pass


class Attachment(object):
    """Encapsulates file attachment information.

    :versionadded: 0.3.5

    :param filename: filename of attachment
    :param content_type: file mimetype
    :param data: the raw file data
    :param disposition: content-disposition (if any)
    """

    def __init__(self, filename=None, content_type=None, data=None,
                 disposition=None, headers=None):
        self.filename = filename
        self.content_type = content_type
        self.data = data
        self.disposition = disposition or 'attachment'
        self.headers = headers or {}


class Message(object):
    """Encapsulates an email message.

    :param subject: email subject header
    :param recipients: list of email addresses
    :param body: plain text message
    :param html: HTML message
    :param sender: email sender address, or **MAIL_DEFAULT_SENDER** by default
    :param cc: CC list
    :param bcc: BCC list
    :param attachments: list of Attachment instances
    :param reply_to: reply-to address
    :param date: send date
    :param charset: message character set
    :param extra_headers: A dictionary of additional headers for the message
    :param mail_options: A list of ESMTP options to be used in MAIL FROM command
    :param rcpt_options:  A list of ESMTP options to be used in RCPT commands
    """

    def __init__(self, subject='',
                 recipients=None,
                 body=None,
                 html=None,
                 sender=None,
                 cc=None,
                 bcc=None,
                 attachments=None,
                 ascii_attachments=None,
                 reply_to=None,
                 date=None,
                 charset=None,
                 extra_headers=None,
                 mail_options=None,
                 rcpt_options=None):

        if isinstance(sender, tuple):
            sender = "%s <%s>" % sender

        self.recipients = recipients or []
        self.subject = subject
        self.sender = sender
        self.reply_to = reply_to
        self.cc = cc or []
        self.bcc = bcc or []
        self.body = body
        self.html = html
        self.date = date
        self.msgId = make_msgid()
        self.charset = charset
        self.extra_headers = extra_headers
        self.mail_options = mail_options or []
        self.rcpt_options = rcpt_options or []
        self.attachments = attachments or []
        if ascii_attachments is None:
            ascii_attachments = current_app.config.get("MAIL_ASCII_ATTACHMENTS", False)
        self.ascii_attachments = ascii_attachments

    @property
    def send_to(self):
        return set(self.recipients) | set(self.bcc or ()) | set(self.cc or ())

    def _mimetext(self, text, subtype='plain'):
        """Creates a MIMEText object with the given subtype (default: 'plain')
        If the text is unicode, the utf-8 charset is used.
        """
        charset = self.charset or 'utf-8'
        return MIMEText(text, _subtype=subtype, _charset=charset)

    def _message(self):
        """Creates the email"""
        ascii_attachments = self.ascii_attachments
        encoding = self.charset or 'utf-8'

        attachments = self.attachments or []

        if len(attachments) == 0 and not self.html:
            # No html content and zero attachments means plain text
            msg = self._mimetext(self.body)
        elif len(attachments) > 0 and not self.html:
            # No html and at least one attachment means multipart
            msg = MIMEMultipart()
            msg.attach(self._mimetext(self.body))
        else:
            # Anything else
            msg = MIMEMultipart()
            alternative = MIMEMultipart('alternative')
            alternative.attach(self._mimetext(self.body, 'plain'))
            alternative.attach(self._mimetext(self.html, 'html'))
            msg.attach(alternative)

        if self.subject:
            msg['Subject'] = sanitize_subject(force_text(self.subject), encoding)

        msg['From'] = sanitize_address(self.sender, encoding)
        msg['To'] = ', '.join(list(set(sanitize_addresses(self.recipients, encoding))))

        msg['Date'] = formatdate(self.date, localtime=True)
        # see RFC 5322 section 3.6.4.
        msg['Message-ID'] = self.msgId

        if self.cc:
            msg['Cc'] = ', '.join(list(set(sanitize_addresses(self.cc, encoding))))

        if self.reply_to:
            msg['Reply-To'] = sanitize_address(self.reply_to, encoding)

        if self.extra_headers:
            for k, v in self.extra_headers.items():
                msg[k] = v

        SPACES = re.compile(r'[\s]+', re.UNICODE)
        for attachment in attachments:
            f = MIMEBase(*attachment.content_type.split('/'))
            f.set_payload(attachment.data)
            encode_base64(f)

            filename = attachment.filename
            if filename and ascii_attachments:
                # force filename to ascii
                filename = unicodedata.normalize('NFKD', filename)
                filename = filename.encode('ascii', 'ignore').decode('ascii')
                filename = SPACES.sub(u' ', filename).strip()

            try:
                filename and filename.encode('ascii')
            except UnicodeEncodeError:
                if not PY3:
                    filename = filename.encode('utf8')
                filename = ('UTF8', '', filename)

            f.add_header('Content-Disposition',
                         attachment.disposition,
                         filename=filename)

            for key, value in attachment.headers:
                f.add_header(key, value)

            msg.attach(f)
        if message_policy:
            msg.policy = message_policy

        return msg

    def as_string(self):
        return self._message().as_string()

    def as_bytes(self):
        if PY34:
            return self._message().as_bytes()
        else: # fallback for old Python (3) versions
            return self._message().as_string().encode(self.charset or 'utf-8')

    def __str__(self):
        return self.as_string()

    def __bytes__(self):
        return self.as_bytes()

    def has_bad_headers(self):
        """Checks for bad headers i.e. newlines in subject, sender or recipients.
        RFC5322: Allows multiline CRLF with trailing whitespace (FWS) in headers
        """

        headers = [self.sender, self.reply_to] + self.recipients
        for header in headers:
            if _has_newline(header):
                return True

        if self.subject:
            if _has_newline(self.subject):
                for linenum, line in enumerate(self.subject.split('\r\n')):
                    if not line:
                        return True
                    if linenum > 0 and line[0] not in '\t ':
                        return True
                    if _has_newline(line):
                        return True
                    if len(line.strip()) == 0:
                        return True
        return False

    def is_bad_headers(self):
        from warnings import warn
        msg = 'is_bad_headers is deprecated, use the new has_bad_headers method instead.'
        warn(DeprecationWarning(msg), stacklevel=1)
        return self.has_bad_headers()

    def send(self, connection):
        """Verifies and sends the message."""

        connection.send(self)

    def add_recipient(self, recipient):
        """Adds another recipient to the message.

        :param recipient: email address of recipient.
        """

        self.recipients.append(recipient)

    def attach(self,
               filename=None,
               content_type=None,
               data=None,
               disposition=None,
               headers=None):
        """Adds an attachment to the message.

        :param filename: filename of attachment
        :param content_type: file mimetype
        :param data: the raw file data
        :param disposition: content-disposition (if any)
        """
        self.attachments.append(
            Attachment(filename, content_type, data, disposition, headers))


class _MailMixin(object):

    state = None

    @contextmanager
    def record_messages(self):
        """Records all messages. Use in unit tests for example::

            with mail.record_messages() as outbox:
                response = app.test_client.get("/email-sending-view/")
                assert len(outbox) == 1
                assert outbox[0].subject == "testing"

        You must have blinker installed in order to use this feature.
        :versionadded: 0.4
        """

        if not email_dispatched:
            raise RuntimeError("blinker must be installed")

        outbox = []

        def _record(message, app):
            outbox.append(message)

        email_dispatched.connect(_record)

        try:
            yield outbox
        finally:
            email_dispatched.disconnect(_record)

    def send(self, message):
        """Sends a single message instance. If TESTING is True the message will
        not actually be sent.

        :param message: a Message instance.
        """
        if not message.sender:
            sender = message.sender or self.default_sender
            if isinstance(sender, tuple):
                sender = "%s <%s>" % sender
            message.sender = sender

        with self.connect() as connection:
            message.send(connection)

    def send_message(self, *args, **kwargs):
        """Shortcut for send(msg).

        Takes same arguments as Message constructor.

        :versionadded: 0.3.5
        """

        self.send(Message(*args, **kwargs))

    def connect(self):
        """Opens a connection to the mail host."""
        app = getattr(self, "app", None) or current_app
        state = self.state or Mail.parse_state(app.config)
        try:
            return Connection(state)
        except KeyError:
            raise RuntimeError("The curent application was not configured with Flask-Mail")


class _Mail(_MailMixin):
    def __init__(self, server, username, password, port, use_tls, use_ssl,
                 default_sender, debug, max_emails, suppress,
                 ascii_attachments=False):
        self.server = server
        self.username = username
        self.password = password
        self.port = port
        self.use_tls = use_tls
        self.use_ssl = use_ssl
        self.default_sender = default_sender
        self.debug = debug
        self.max_emails = max_emails
        self.suppress = suppress
        self.ascii_attachments = ascii_attachments


class Mail(_MailMixin):
    """Manages email messaging

    :param app: Flask instance
    """

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.state = self.init_app(app)
        else:
            self.state = None

    @classmethod
    def parse_state(cls, config, debug=False, testing=False):
        return _Mail(
            config.get('MAIL_SERVER', '127.0.0.1'),
            config.get('MAIL_USERNAME'),
            config.get('MAIL_PASSWORD'),
            config.get('MAIL_PORT', 25),
            config.get('MAIL_USE_TLS', False),
            config.get('MAIL_USE_SSL', False),
            config.get('MAIL_DEFAULT_SENDER'),
            int(config.get('MAIL_DEBUG', debug)),
            config.get('MAIL_MAX_EMAILS'),
            config.get('MAIL_SUPPRESS_SEND', testing),
            config.get('MAIL_ASCII_ATTACHMENTS', False)
        )

    def init_mail(self, config, debug=False, testing=False):
        return self.__class__.parse_state(config, debug, testing)

    def init_app(self, app):
        """Initializes your mail settings from the application settings.

        You can use this if you want to set up your Mail instance
        at configuration time.

        :param app: Flask application instance
        """
        state = self.init_mail(app.config, app.debug, app.testing)

        return state

    def __getattr__(self, name):
        return getattr(self.state, name, None)


signals = blinker.Namespace()

email_dispatched = signals.signal("email-dispatched", doc="""
Signal sent when an email is dispatched. This signal will also be sent
in testing mode, even though the email will not actually be sent.
""")


def mail_account(address, app=None):
    mail = Mail()
    app = app or current_app
    config = app.config.copy()
    state = Mail.parse_state(config)
    mails = config.get("MAILS", {})
    account = mails.get(address, {})
    for k, v in account.items():
        setattr(state, k.lower(), v)
    mail.app = app
    mail.state = state
    return mail



# -*- coding: utf-8 -*-
import simplejson as pjson
from json import JSONEncoder
from datetime import datetime, time
from .func import pycallable
from collections import Iterable


class JSON(JSONEncoder):

    @staticmethod
    def default(o):
        if isinstance(o, Iterable):
            return list(o)

        if hasattr(o, '__json__') and pycallable(getattr(o, '__json__')):
            return o.__json__()

        if hasattr(o, '__dict__'):
            return o.__dict__

        if isinstance(o, datetime):
            return o.strftime("%Y-%m-%d %H:%M:%S")

        if isinstance(o, time):
            return o.strftime("%H:%M:%S")

        return "%s" % o

    @classmethod
    def stringify(cls, o, ensure_ascii=False, separators=(',', ':'), encoding='utf-8', **kws):
        return pjson.dumps(o, default=cls.default, use_decimal=True, ensure_ascii=ensure_ascii, separators=separators, encoding=encoding, **kws)

    @classmethod
    def parse(cls, str, **kws):
        return pjson.loads(str, **kws)

    @classmethod
    def dict(cls, o, **kws):
        return cls.parse(cls.stringify(o, **kws))



# -*- coding: utf-8 -*-
#
# 读取文件的hash值
# from wesell import filehash
# fmd5 = filehash.md5(myfile_path)
#

import hashlib

__buffsize = 64 * 1024


def __hash(method, *args):
    calc = getattr(hashlib, method.__name__)

    def _c(file):
        with open(file, 'rb') as f:
            sha1obj = calc()
            while True:
                data = f.read(__buffsize)
                if not data:
                    break
                sha1obj.update(data)
            hashstr = sha1obj.hexdigest()
            return hashstr
    return _c


@__hash
def sha1(file):
    pass


@__hash
def sha224(file):
    pass

@__hash
def sha256(file):
    pass

@__hash
def sha384(file):
    pass


@__hash
def sha512(file):
    pass


@__hash
def md5(file):
    pass


# -*- coding: utf-8 -*-
# python2 / python3 exception 兼容
# 使用:
# from wesell.exceptions import *
from __future__ import absolute_import
import sys

if sys.version_info.major == 2:
    from exceptions import *




# -*- coding: utf-8 -*-

from .globals import *
from flask import request, make_response, has_request_context
from sqlalchemy import event
from sqlalchemy.orm import mapper
from sqlalchemy.inspection import inspect
from .. import isint_or, JSON, pycallable
import base64
import random
import re
import time
import sys


def apply_app(app):
    stacker["app"] = app
    app.config.setdefault("CACHE_LIFE_TIME", 600)


class BaseInterface(object):
    """基础接口类"""
    # 绑定指定模型类, 可继承模型的类方法(@classmethod)
    __bindmodel__ = None

    def __getattr__(self, name):
        if self.__bindmodel__ is not None:
            return getattr(self.__bindmodel__, name, None)
        return None


class version_control:
    def __getattribute__(self, name):  # real signature unknown; restored from __doc__
        """ x.__getattribute__('name') <==> x.name """
        pass

    def __get__(self, obj, type=None):  # real signature unknown; restored from __doc__
        """ descr.__get__(obj[, type]) -> value """
        pass

    def __set__(self, obj, type=None):  # real signature unknown; restored from __doc__
        """ descr.__get__(obj[, type]) -> value """
        pass

    def __del__(self, obj, type=None):  # real signature unknown; restored from __doc__
        """ descr.__get__(obj[, type]) -> value """
        pass

    def __init__(self, function):  # real signature unknown; restored from __doc__
        pass

    __func__ = property(__get__, __set__, __del__)  # default


def version(ver):
    """版本号修饰"""
    return version_control


def operator_role():
    """请求头消息中获取操作者数据"""
    return request.headers.get("Operator-Role", "") if has_request_context() else ""


def operator_id():
    """请求头消息中获取操作者数据"""
    return request.headers.get("Operator-Id", 0) if has_request_context() else 0


def operator_name():
    """请求头消息中获取操作者数据"""
    return request.headers.get("Operator-Name", "") if has_request_context() else ""


def instant_defaults_listener(target, args, kwargs):
    """
    模型初始化时, 字段使用默认值
    """
    for key, column in inspect(target.__class__).columns.items():
        # if key in ["creatorid", "operatorid"]:
        #     # 请求头消息中获取操作者数据, 自动填入creatorid, operatorid字段
        #     operatorid = operator_id()
        #     if operatorid is not None:
        #         setattr(target, key, operatorid)
        #         continue
        # # operator role
        # if key in ["creator_role", "operator_role"]:
        #     # 请求头消息中获取操作者数据, 自动填入creatorid, operatorid字段
        #     role = operator_role()
        #     if operatorrole is not None:
        #         setattr(target, key, role)
        #         continue

        if column.default is not None:
            if pycallable(column.default.arg):
                setattr(target, key, column.default.arg(target))
            else:
                setattr(target, key, column.default.arg)

# 默认值初始化
event.listen(mapper, 'init', instant_defaults_listener)


def redirect(location, code=302, autocorrect=False):
    response = make_response()
    response.status_code = code
    response.autocorrect_location_header = autocorrect
    response.headers['Location'] = location
    return response


def response_json(data, **kwargs):
    """
    响应json数据
    :param data:
    :param kwargs:
    :return:
    """
    response = make_response()
    json = JSON.stringify(data, **kwargs)
    # ---------------
    # Cache-Control
    # ---------------
    # HTTP/1.1
    response.headers["Cache-Control"] = "no-cache, no-store, max-age=0, must-revalidate"
    # HTTP/1.0
    response.headers["Pragma"] = "no-cache"
    # Expires
    response.headers["Expires"] = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())

    response.headers["Content-Type"] = "text/html; charset=utf-8"
    # jsvar 模式
    jsvar = request.args.get("jsvar", None)
    if jsvar:
        response.data = "<script>%s=%s</script>" % (jsvar, json)
        return response
    # script 标签回调
    scriptcallback = __strict_callback(request.args.get("scriptcallback", ""))
    if scriptcallback:
        response.data = "<script>%s(%s)</script>" % (scriptcallback, json)
        return response
    # jsonpcallback, 兼容参数callback
    jsonp = request.args.get("callback", "")
    jsonp = jsonp if jsonp else request.args.get("jsonpcallback", "")
    jsonp = __strict_callback(jsonp)
    if jsonp:
        response.headers["Content-Type"] = "application/javascript; charset=utf-8"
        response.data = "%s && %s(%s)" % (jsonp, jsonp, json)
    # 常规
    response.headers["Content-Type"] = "application/json; charset=utf-8"
    boundary = request.headers.get("Boundary-Wrap", "")
    if boundary.lower() == "no":
        boundary = ""
    if boundary.lower() == "passive":
        boundary == base64.b64encode("%s" % random.random())
    response.headers["Boundary"] = boundary
    response.data = "%s%s%s" % (boundary, json, boundary)
    return response


def __strict_callback(callback):
    pattern = re.compile("[^\w\._]")
    return callback if pattern.match(callback) else ""



# -*- coding: utf-8 -*-
import requests
from future.utils import with_metaclass
from hashlib import md5
from inspect import getargspec
from functools import wraps
from . import JSON
from . import app
import sys

if sys.version_info.major == 2:
    from exceptions import *

# 单实例缓存
__instances__ = {}


def client_instance(cls):
    """ client 单例缓存
    :param cls:
    :return:
    """
    path = ".".join([cls.__module__, cls.__name__])
    ins = __instances__
    if path in __instances__:
        return __instances__.get(path)
    __instances__[path] = cls()
    return __instances__.get(path)


def description(method):
    """方法描述说明, 用于生成注释文档

    - dsf
    - sdfaf
    """

    @classmethod
    @wraps(method)
    def execute(cls, *args, **kwargs):
        argspec = getargspec(method)
        argnames = argspec.args[1:] if len(argspec.args) else ()  # tuple
        argdef_list = argspec.defaults if argspec.defaults else ()  # tuple
        argdefs = dict()  # 参数默认值, 字典, 保存有默认值的参数名和值
        defindex = len(argnames) - len(argdef_list)
        args = list(args)
        for i, defv in enumerate(argdef_list):
            # 绑定默认值
            argdefs[argnames[defindex + i]] = defv

        for i, name in enumerate(argnames):
            if not kwargs.has_key(name):
                if len(args):
                    kwargs[name] = args.pop(0)
                else:
                    kwargs[name] = argdefs[name]

        return client_instance(cls).execute(method.func_name, *args, **kwargs)

    return execute


class ClientMeta(type):
    """ Client Meta Class
    """

    def __getattr__(cls, method):
        return getattr(client_instance(cls), method, None)

    def __str__(cls):
        return 'Interface Client <%s>' % (cls.__name__,)


class BaseClient(with_metaclass(ClientMeta, object)):
    """ Remote Interface Base Client
        创建远程接口客户端, 用于请求远程接口
    """
    # 远程接口名, 默认为Client的类名去掉"Client"
    __interface__ = None
    # 远程接口API, 可在configs中配置, 配置名为 "CLIENT_%s_API" % self.__interface__
    # 如 http://127.0.0.1:9001/rpc
    __api__ = None

    def __init__(self, interface=None, api=None):
        if interface is None:
            interface = self.__default_interface()
        # class 中未定义__interface__
        if not self.__interface__:
            self.__interface__ = interface
        # __api__ 需要依赖 __interface__
        if api is not None:
            self.__api__ = api
        self.__api__ = self.__default_api()

    def api(self):
        return self.__api__

    def __default_interface(self):
        cn = self.__class__.__name__
        return cn[0:-6]

    def __default_api(self):
        if self.__api__:
            return self.__api__
        api = self.__config_api()
        if not api:
            api = self.__parent_config_api()
        return api

    def __config_api(self):
        if self.__interface__ is None:
            return None
        name = "CLIENT_%s_API" % self.__interface__.upper()
        return app.config.get(name, None)

    def __parent_config_api(self):
        """ 尝试使用父类在configs中的API配置
        :return:
        """
        cls = self.__class__
        if cls.__name__ == "BaseClient":
            return None
        bases = cls.__bases__
        for base in bases:
            inst = client_instance(base)
            if hasattr(inst, "__config_api"):
                api = inst.__config_api()
                if api:
                    return api
                else:
                    return inst.__parent_config_api()
        return None

    def execute(self, method, *args, **kwargs):
        """ 接口执行方法
        """
        raise NotImplementedError()

    def __getattr__(self, method):
        if method.startswith("__"):
            raise AttributeError("%s has no attribute: %s" % (self, method))

        def callmethod(*args, **kwargs):
            return self.execute(method, *args, **kwargs)

        return callmethod

    pass


class PHPV1Client(BaseClient):
    """ php老接口, 对应if.wesell.org

    """

    version = "0.0.1"

    def __init__(self, *args, **kwargs):
        BaseClient.__init__(self, *args, **kwargs)
        if not self.__api__:
            path = None
            if hasattr(self, "__module_path__"):
                path = self.__module_path__
            else:
                path = self.__interface__
            self.__api__ = "http://if.wesell.org/%s/index.php" % (path.lower())

    def execute(self, method, *args, **kwargs):
        """ 接口执行方法
        """
        if not self.__api__:
            raise Exception("Undefined API for %s" % self.__class__.__name__)
        # 参数数据
        params = dict(
            mod=self.__interface__,
            act=method,
            platform="diandao"
        )
        params.update(**kwargs)
        randkey = []
        pkeys = params.keys()
        pkeys.sort()
        key = pkeys[-1]
        randkey.append("%s%s" % (key, params[key]))
        # 添加干扰
        randkey = "".join(randkey) + "The key for Programming"
        # md5
        randkey = md5(randkey).hexdigest()
        # 前9位字符
        randkey = randkey[0:9]
        # http请求参数
        data = dict(
            randkey=randkey,
            c_version=self.version,
            parameter=JSON.stringify(params)
        )
        res = requests.get(self.api(), data)

        if not res.ok:
            raise Exception(res.reason, res.status_code)

        result = res.json()
        if result['code'] > 0:
            raise Exception("[Error] code:%s; desc:%s" % (result['code'], result['desc']))
        return result['data']


class ApiBaseClient(BaseClient):
    version = "0.0.1"

    def __init__(self, *args, **kwargs):
        BaseClient.__init__(self, *args, **kwargs)
        if not self.__api__:
            path = None
            if hasattr(self, "__module_path__"):
                path = self.__module_path__
            else:
                path = self.__interface__
            self.__api__ = "http://api.wesell.org/%s" % path

    def execute(self, method, *args, **kwargs):
        """ 接口执行方法
        """
        if not self.__api__:
            raise Exception("Undefined API for %s" % self.__class__.__name__)
        # 参数数据
        params = dict()
        params.update(**kwargs)
        res = requests.get("%s/%s" % (self.api(), method), params)
        if not res.ok:
            raise Exception(res.reason, res.status_code)
        result = res.json()
        if result['error']:
            raise Exception("[Error] code:%s; desc:%s" % (result['error']['code'], result['error']['message']))
        return result['data']





# -*- coding: utf-8 -*-

from flask import Blueprint, request
import interfaces
from . import JSON, pycallable
import sys, inspect
import traceback

blue = Blueprint('rpc', 'rpc')

# interface caches
_caches = {}


@blue.route('/rpc/<string:interface>/<string:method>', methods=["POST", "GET"])
def rpc(interface, method):
    if not interface or not method:
        return error_404(interface, method)
    interface = "%sInterface" % interface
    intf = None
    # 缓存接口实例, 避免多次创建
    if hasattr(_caches, str(interface).lower()):
        intf = _caches[str(interface).lower()]
    if intf is None and hasattr(interfaces, interface):
        intf = getattr(interfaces, interface)()
        _caches[str(interface).lower()] = intf

    # 无接口类
    if intf is None:
        return error_404(interface, method)

    # 接口无指定方法
    if not hasattr(intf, method):
        return error_404(interface, method)

    callee = getattr(intf, method)

    if not pycallable(callee):
        return error_404(interface, method)

    params = []
    data = None
    error = None

    try:
        params = get_params()
    except ValueError as e:
        # params参数解析异常
        error = {
            "code": 501,
            "message": "params parse error:%s" % e
        }
        return rpc_response(data, error=error)

    try:
        # 数组参数
        if isinstance(params, list):
            data = callee(*params)
        # kwargs 参数
        if isinstance(params, dict):
            data = callee(**params)
    except Exception as e:
        info = sys.exc_info()
        errtype = info[0].__name__
        message = info[1].message
        trace = get_exc_traceback(info[2])

        error = {
            "code": 500,
            "message": "%s: %s on %s at line %s" % (errtype, message, trace['file'], trace['line'])
        }
    finally:
        return rpc_response(data, error=error)


def rpc_response(data, error=None, status=200):
    return JSON.stringify({
        "status": status,
        "error": error,
        "data": data
    })


def get_params():
    """获取请求中的params参数"""
    ps = request.values.get('params', type=str, default="[]")
    ps = JSON.parse(ps)
    return ps


def error_404(interface='', method=''):
    error = {
        "code": 404,
        "message": "Undefined Interface:%s.%s()" % (interface, method)
    }
    return rpc_response(None, error=error, status=404)


def get_exc_traceback(trace):
    if hasattr(trace, "tb_next") and trace.tb_next:
        return get_exc_traceback(trace.tb_next)

    trace = {
        "line": trace.tb_lineno,
        "file": inspect.getfile(trace.tb_frame)
    }
    return trace



# -*- coding: utf-8 -*-
from pkgutil import walk_packages
from importlib import import_module
from flask import Flask, request
from . import stacker, redis, db, celery, rq
from flask import _app_ctx_stack
from inspect import ismodule
from xloger.vendors import FlaskXLoger
from sys import exc_info
import inspect
# from deprecation import deprecated
import logging
import time

DEFAULT_APP_NAME = "application"

_vars = {"app": None, "config": {}, "db": None}


def get_exc_traceback(trace):
    if hasattr(trace, "tb_next") and trace.tb_next:
        return get_exc_traceback(trace.tb_next)

    trace = {
        "line": trace.tb_lineno,
        "file": inspect.getfile(trace.tb_frame)
    }
    return trace


def create_app(config, app_name=None):
    if app_name is None:
        app_name = DEFAULT_APP_NAME
        if isinstance(config, object):
            app_name = getattr(config, "APP_NAME", DEFAULT_APP_NAME)
        
    app = Flask(app_name)
    if config is None:
        return app
    _vars['app'] = app
    _app_ctx_stack.push(app.app_context())
    app.url_map.strict_slashes = app.config.get("URL_MAP_STRICT_SLASHES", False)
    configure_app(app, config)
    configure_db(app)
    configure_redis(app)
    # configure_celery(app)
    # configure_rq(app)
    # configure_models(app)
    FlaskXLoger(app, disabled=not app.config.get("XLOGER_ENABLED", False))
    #configure_blueprints(app)
    configure_handler(app)
    return app


def configure_handler(app):
    gunicorn_error_handlers = logging.getLogger('gunicorn.error').handlers
    app.logger.handlers.extend(gunicorn_error_handlers)

    @app.errorhandler(Exception)
    def server_internal_error(e):
        info = exc_info()
        message = e.__repr__()
        trace = get_exc_traceback(info[2])
        if app.config.get("DEBUG"):
            message = "%s in %s on line %d" % (message, trace['file'], trace['line'])
        app.logger.error(message)
        return message, 500


# ------------
# 应用配置文件
def configure_app(app, config=None):
    # config 不为空时覆盖配置
    if config is not None:
        app.config.from_object(config)
    # 映射配置
    _vars['config'] = app.config
    stacker['app'] = app


def configure_db(app):
    """
    数据库连接初始化

    :config SQLALCHEMY_POOL_RECYCLE: 连接池生命周期
    :config SQLALCHEMY_POOL_TIMEOUT: 连接池超时
    :config SQLALCHEMY_BINDS: 数据库连接配置, 可配置多个key=>value的配置

    :param app:
    :return:
    """
    app.config.setdefault("CACHE_LIFE_TIME", 600)
    db.init_app(app)


def configure_redis(app):
    """
    Redis 配置, 需要configs.py中的Redis参数

    :config REDIS: (host, port, database)

    :param app:
    :return:
    """
    max_connections = app.config.get("REDIS_MAX_CONNECTIONS", None)
    redis.init_app(app, max_connections=max_connections)


def configure_celery(app):
    celery.main = app.name
    celery.conf.update(app.config.get("CELERY", dict()))
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    app.celery = celery


def configure_rq(app):
    rq.init_app(app)


# def configure_clients(app):
#     try:
#         import clients
#     except ImportError:
#         return
#
#     for loader, name, is_pkg in walk_packages(clients.__path__):
#         import_module(clients.__name__ + '.' + name)
#         client_module = getattr(clients, name)
#         setattr(clients, name, getattr(client_module, name))


# 注册蓝图(Route)
def configure_blueprints(app):
    import blueprints

    @app.before_request
    def before_request():
        request.timestamp = time.time()

    prints = []
    for loader, name, is_pkg in walk_packages(blueprints.__path__):
        import_module(blueprints.__name__ + '.' + name)

    for name, blue in blueprints.__dict__.items():
        if not ismodule(blue):
            continue
        if hasattr(blue, "blueprint"):
            prints.append((blue.blueprint, getattr(blue, "url_prefix", None)))

    for blue, url_prefix in prints:
        app.register_blueprint(blue, url_prefix=url_prefix)


def configure_models(app):
    import models
    for loader, name, is_pkg in walk_packages(models.__path__):
        import_module(models.__name__ + '.' + name)
        module = getattr(models, name)
        model = getattr(module, name, None)
        if model is None:
            raise ImportError("class %s not found in %s" % (name, module.__file__))
        setattr(models, name, model)


# 获取app配置
def get_config():
    return _vars['config']


# 获取app配置
def get_app():
    return _vars['app']


def get_db():
    return db


# 多数据库时, 取到指定数据库连接
def get_db_engine(bind):
    db = get_db()
    app = get_app()
    return db.get_engine(app, bind)


# 初始化数据库
def init_db():
    """
    数据库初始化
    - 未创建表的模型会自动创建数据表
    - 已存在的表, 不会自动更新字段信息

    :return:
    """
    db = get_db()
    db.create_all()



from setuptools import setup

setup(
    name='wesell',
    version="1.1.1",
    url='https://gitee.com/idollo/wesell.pylib.git',
    license='MIT',
    author='idollo',
    author_email='stone58@qq.com',
    description='wesell flask python libs.',
    long_description=__doc__,
    packages=['wesell', 'wesell.flask', 'wesell.sqlalchemy'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'future>=0.17.1',
        'Flask==0.12.2',
        'SQLAlchemy==1.2.9',
        'Flask-Script>=2.0.5',
        'Flask==0.12.2',
        'Flask-Celery==2.4.3',
        'Flask-HTTPAuth==3.2.3',
        'Flask-Redis==0.3.0',
        'Flask-RQ==0.2',
        'Flask-Session==0.3.1',
        'Flask-SQLAlchemy==2.1',
        'pymysql==0.9.3',
        'mysql>=0.0.2',
        'mysqlclient>=1.3.0,<1.4'
        'redis>=2.10.5',
        'celery==4.3.0',
        'hiredis==1.0.0',
        'msgpack>=0.6.2',
        'xloger>=1.6.5',
        'simplejson>=3.11.1',
        'dill>=0.3.3'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ]
)
